/**
 * Created on Sep 11, 2003
 * @author Richard Gradischnegg RG
 */
package isac.bridge;

import isac.interfaces.ICalcElement;
import isac.interfaces.ICalcIterator;
import isac.util.formulae.Assumptions;
import isac.util.formulae.CalcHead;
import isac.util.formulae.Context;
import isac.util.formulae.ContextMethod;
import isac.util.formulae.ContextProblem;
import isac.util.formulae.ContextTheory;
import isac.util.formulae.HierarchyKey;
import isac.util.formulae.Match;
import isac.util.formulae.MethodID;
import isac.util.formulae.Position;
import isac.util.formulae.ProblemID;
import isac.util.tactics.Tactic;
import isac.wsdialog.IContextProvider.ContextType;

import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Vector;

import org.apache.log4j.Logger;

/**
 * Iterator class to navigate through a calcTree. CalcIterator implements
 * several move Methods to navigate up/down a CalcTree, go into Subproblems,..
 * One can retrive also the element in the CalcTree currently marked by this
 * Iterator.
 * 
 * @author Richard Gradischnegg
 */
public class CalcIterator implements ICalcIterator, Serializable {

    private static final Logger logger = Logger.getLogger(CalcIterator.class
            .getName());

    private CalcTree calc_tree_;

    private int calc_tree_id_;// from SML + mapped by

    private static final long serialVersionUID = 36493649573845958L;

    // smlCalcIDtoJavaCalcID(int);

    // iterator_id_ from SML TODO.WN041116 drop with new Iterator, which is NOT
    // stored in sml-kernel any more
    private int iterator_id_;

    // WN041209 i.e. + SML: make 'Iterator n 1' implicit to 'CalcTree n' ?

    private Position sml_pos_;// unique within the sml calctree

    /**
     * Construct a new calculation element iterator object attached to the
     * passed calculation tree
     * 
     * @param calcTree
     *            calculation tree
     */
    protected CalcIterator(CalcTree calcTree) {
        if (logger.isDebugEnabled())
            logger.debug("CalcIterator(new Position)");
        this.setSmlPos(new Position());
        this.calc_tree_ = calcTree;
        this.calc_tree_id_ = calcTree.getId();
        // try {
        // this.iterator_id_ = calcTree.getMathEngine().iterator(calc_tree_id_);
        // } catch (RemoteException e) {
        //		
        // e.printStackTrace();
        // }
        this.iterator_id_ = 4711;// this selects for changing the CalcTree
    }

    /**
     * Construct a new calculation element iterator object attached to the
     * passed calculation tree with initial position (fetched from a formula on
     * the worksheet)
     * 
     * @param calcTree
     *            calculation tree
     */
    public CalcIterator(CalcTree calcTree, Position sml_pos) {
        if (logger.isDebugEnabled())
            logger.debug("CalcIterator(" + sml_pos.toSMLString() + ")");
        this.setSmlPos(sml_pos);
        this.calc_tree_ = calcTree;
        this.calc_tree_id_ = calcTree.getId();
        this.iterator_id_ = 4711;// this is calculated by the KE without
                                    // state
        // try {
        // this.iterator_id_ = calcTree.getMathEngine().iterator(calc_tree_id_);
        // } catch (RemoteException e) {
        //			
        // e.printStackTrace();
        // }
    }

    /**
     * Turn this iterator into a HotSpot iterator i.e. use it to move the active
     * Formula. This operation is not reversible.
     * 
     * @deprecated since autoCalculate etc. is applied to _CalcTree_ and all
     *             other CalcIterators just read
     */
    void makeHotSpot() {
        this.iterator_id_ = 1;
    }

    /**
     * Reference the first (root) element in the attached calculation tree.
     * 
     * @return Like all the moveXXX methods this returns true on success, false
     *         if there is no element to move to.
     */
    public boolean moveRoot() {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: moveRoot()");
        try {
            // WN041208 return
            // calc_tree_.getMathEngine().moveRoot(calc_tree_id_,
            // iterator_id_);
            Position p = calc_tree_.getMathEngine().moveRoot(calc_tree_id_,
                    iterator_id_);
            if (p == null) {
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveRoot() <- false NULL###");
                return false;
            } else {
                this.setSmlPos(p);
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveRoot() <- true");
                return true;
            }
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: moveRoot() <- false RemoteException###");
        return false;
    }

    /**
     * Reference the previous formula in the attached calculation tree.
     * 
     * @return Like all the moveXXX methods this returns true on success, false
     *         if there is no element to move to.
     */
    public boolean moveUp() {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: moveUp()");
        try {
            // WN041208 return calc_tree_.getMathEngine().moveUp(calc_tree_id_,
            // iterator_id_, sml_pos_);
            Position p = calc_tree_.getMathEngine().moveUp(calc_tree_id_,
                    iterator_id_, sml_pos_);
            if (p == null) {
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveUp() <- false NULL###");
                return false;
            } else {
                this.setSmlPos(p);
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveUp() <- true");
                return true;
            }
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: moveUp() <- false RemoteException###");
        return false;
    }

    /**
     * Reference the next formula in the attached calculation tree.
     * 
     * @return Like all the moveXXX methods this returns true on success, false
     *         if there is no element to move to.
     */
    public boolean moveDown() {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: moveDown()");
        try {
            // WN041208 return
            // calc_tree_.getMathEngine().moveDown(calc_tree_id_,
            // iterator_id_, sml_pos_);
            Position p = calc_tree_.getMathEngine().moveDown(calc_tree_id_,
                    iterator_id_, sml_pos_);
            if (p == null) {
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveDown() <- false NULL###");
                return false;
            } else {
                this.setSmlPos(p);
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveDown() <- true");
                return true;
            }
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: moveDown() <- false RemoteException###");
        return false;
    }

    /**
     * Enter the next coarser level of refinement in the attached calculation
     * tree.
     * 
     * @return Like all the moveXXX methods this returns true on success, false
     *         if there is no element to move to.
     */
    public boolean moveLevelUp() {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: moveLevelUp()");
        try {
            // WN041208 return
            // calc_tree_.getMathEngine().moveLevelUp(calc_tree_id_,
            // iterator_id_,
            // sml_pos_);
            Position p = calc_tree_.getMathEngine().moveLevelUp(calc_tree_id_,
                    iterator_id_, sml_pos_);
            if (p == null) {
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveLevelUp() <- false NULL###");
                return false;
            } else {
                this.setSmlPos(p);
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveLevelUp() <- true");
                return true;
            }
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: moveLevelUp() <- false RemoteException###");
        return false;
    }

    /**
     * Enter the next more detailed level of refinement in the attached
     * calculation tree.
     * 
     * @return Like all the moveXXX methods this returns true on success, false
     *         if there is no element to move to.
     */
    public boolean moveLevelDown() {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: moveLevelDown()");
        try {
            // WN041208 return
            // calc_tree_.getMathEngine().moveLevelDown(calc_tree_id_,
            // iterator_id_, sml_pos_);
            Position p = calc_tree_.getMathEngine().moveLevelDown(
                    calc_tree_id_, iterator_id_, sml_pos_);
            if (p == null) {
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveLevelDown() <- false NULL###");
                return false;
            } else {
                this.setSmlPos(p);
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveLevelDown() <- true");
                return true;
            }
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: moveLevelDown() <- false RemoteException###");
        return false;
    }

    /**
     * Reference the calc_head_ that belongs to the currently selected item
     */
    public boolean moveCalcHead() {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: moveCalcHead()");
        try {
            // WN041208return
            // calc_tree_.getMathEngine().moveCalcHead(calc_tree_id_,
            // iterator_id_, sml_pos_);
            Position p = calc_tree_.getMathEngine().moveCalcHead(calc_tree_id_,
                    iterator_id_, sml_pos_);
            if (p == null) {
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveCalcHead() <- false NULL###");
                return false;
            } else {
                this.setSmlPos(p);
                if (logger.isInfoEnabled())
                    logger.info("  DG<-BR: moveCalcHead() <- true");
                return true;
            }
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: moveCalcHead() <- false RemoteException###");
        return false;
    }

    /**
     * Reference the tactic which resulted in the currently selected element of
     * the attached calculation tree.
     * <p>
     * A special case arises if the tactic to continue the current calculation
     * is solving a subproblem. The subproblem is a calculation on its own,
     * represented by a separate calculation(sub-)tree object. A reference to
     * the calculationtree representing the subproblem is passed to the
     * worksheet. The worksheet will handle the situation accordingly, e.g. by
     * opening a new worksheet or displaying the new calculationtree (indented)
     * in the current worksheet.
     * 
     * @return Like all the moveXXX methods this returns true on success, false
     *         if there is no element to move to.
     * @deprecated
     */
    public boolean moveTactic() {
        try {
            return calc_tree_.getMathEngine().moveTactic(calc_tree_id_,
                    iterator_id_);
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        return false;
    }

    /**
     * Reference the formula which resulted in the currently selected element of
     * the attached calculation tree.
     * 
     * @deprecated
     */
    public boolean moveFormula() {
        try {
            return calc_tree_.getMathEngine().moveFormula(calc_tree_id_,
                    iterator_id_);
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        return false;
    }

    /*
     * public Formula getFormula() { return new Formula(); }
     * 
     * public Tactic getTactic() { return new Tactic(); }
     */

    /**
     * The current level of nesting into subcalculations.
     * 
     * @return An int representing level of nesting (higher == more detailed)
     *         <p>
     *         Special values:
     *         <p>
     *         0: outmost level: only first and last step of calculation are on
     *         this level
     */
    public int getLevel() {
        int lev;
        Position pos = this.getPosition();
        if (pos != null) {
            lev = pos.getIntList().size();
            return lev;
        } else {
            return -1;
        }
    }

    /**
     * Returns the item that this iterator is currently referencing
     * 
     * @see isac.wsdialog.ICalcIterator#getElement()
     * @deprecated in favour of get
     */
    public ICalcElement getElement() {

        ICalcElement result = null;
        try {
            result = calc_tree_.getMathEngine().getElement(calc_tree_id_,
                    sml_pos_);
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        return result;
    }

    /**
     * @see isac.interfaces.IToCalc#getFormulaeFromTo(isac.interfaces.ICalcIterator,
     *      isac.interfaces.ICalcIterator, java.lang.Integer, boolean)
     */
    public Vector getFormulaeFromTo(ICalcIterator iterator_to, Integer level,
            boolean result_includes_tactics) throws RemoteException {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: getFormulaeFromTo(" + this.toSMLString() + ", "
                + iterator_to.toSMLString() + level + ")");
        try {
            return calc_tree_.getMathEngine().getFormulaeFromTo(calc_tree_id_,
                    this, iterator_to, level, result_includes_tactics);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * Set the position of this iterator to the position of another, given
     * iterator
     * 
     * @param calcIterator
     *            the other iterator
     * @deprecated WN041208 for reading sml_pos from kernel
     *             (clone,moveActiveFormula)
     */
    void moveTo(CalcIterator calcIterator) {
        this.moveRoot();
        Position otherPos = calcIterator.getPosition();
        // System.out.println(this.getPosition().toSMLString());
        // System.out.println(calcIterator.getPosition().toSMLString());
        while (!this.isOnPosition(otherPos)) {
            if (!this.moveDown())
                return;
            // System.out.println("clone:
            // this=="+this.getPosition().toSMLString());
            // System.out.println("clone:other=="+calcIterator.getPosition().toSMLString());
        }
    }

    // get the Position of the current element in the CalcTree
    public Position getPosition() {// TODO.WN041208 getSmlPos
        if (logger.isInfoEnabled())
            logger.info("  DG<>BR: getPosition() <--" + sml_pos_.toSMLString());
        return sml_pos_;
        // Position thisPos = null;
        // ICalcElement ce = this.getElement();
        // if (ce instanceof CalcHead) {
        // thisPos = ((CalcHead) ce).getPosition();
        // } else if (ce instanceof Formula) {
        // thisPos = ((Formula) ce).getPosition();
        // }
        // return thisPos;
    }

    private void setSmlPos(Position p) {
        this.sml_pos_ = p;
    }

    /**
     * check whether the iterator is on a given position
     * 
     * @deprecated
     */
    private boolean isOnPosition(Position pos) {
        Position thisPos = this.getPosition();
        if (thisPos == null || pos == null)
            return false;
        if (!thisPos.equals(pos))
            return false;
        return true;
    }

    /**
     * Clone this iterator. The result is a second, different iterator
     * referencing the same item as this iterator
     * 
     * @see isac.wsdialog.ICalcIterator#clone()
     */
    public ICalcIterator cloneIterator() {
        if (logger.isDebugEnabled())
            logger.debug("CalcIterator.clone() <-" + sml_pos_.toSMLString());
        CalcIterator clone = (CalcIterator) calc_tree_.iterator();
        clone.setSmlPos(sml_pos_);
        return clone;
    }

    /**
     * Compare the referenced item of this iterator to the item of a second,
     * given iterator
     * 
     * @return Integer
     *         <p>
     *         0 if the second iterator references the same item
     *         <p>
     *         -1 otherwise
     * @see java.lang.Comparable#compareTo(java.lang.Object)
     * 
     */
    public int compareToIterator(ICalcIterator o) {
        int comp = this.sml_pos_.compareTo(((CalcIterator) o).sml_pos_);
        if (logger.isInfoEnabled())
            logger.info("  DG<>BR: compareTo() <- " + comp);
        return comp;
    }

    public boolean equals(Object o) {
        return (this.compareToIterator((ICalcIterator) o) == 0);
    }

    // public int hashCode() {
    // return getPosition().toSMLString().hashCode();
    // }

    public int getIteratorID() {
        return iterator_id_;
    }

    /*
     * (non-Javadoc)
     * 
     * @see isac.interfaces.ICalcIterator#isLast()
     * @deprecated since CalcChanged.last_generated_formula
     */
    public boolean isLast() {
        // TODO: implement
        return true;
    }

    public boolean onCalcHead() throws RemoteException {
        return (sml_pos_.getKind().equals("Pbl") || sml_pos_.getKind().equals(
                "Met"));
    }

    /**
     * @see isac.interfaces.ICalcIterator#onEndOfCalculation()
     */
    public boolean onEndOfCalculation() throws RemoteException {
        return sml_pos_.onEndOfCalculation();
    }

    /**
     * @see isac.interfaces.ICalcIterator#switchToGuard()
     */
    public boolean switchToGuard() throws RemoteException {
        if (onCalcHead()) {
            sml_pos_.setKind("Met");
            return true;
        } else
            return false;
    }

    /**
     * @see isac.interfaces.ICalcIterator#switchToModel()
     */
    public boolean switchToModel() throws RemoteException {
        if (onCalcHead()) {
            sml_pos_.setKind("Pbl");
            return true;
        } else
            return false;
    }

    /**
     * get the formula-part at a CalcIterator-position this can be either a
     * Formula or a CalcHead
     * 
     */
    public ICalcElement getFormula() {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: getFormula()");
        ICalcElement fch = null;
        try {
            fch = calc_tree_.getMathEngine()
                    .getElement(calc_tree_id_, sml_pos_);
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: getFormula <- formula=" + fch.toSMLString());
        return fch;
    }

    /**
     * @see isac.interfaces.ICalcIterator#getTactic()
     */
    public Tactic getTactic() { // WN050217 kopiert von getFormula
        // !!!!!!!!!!!!!!!!!
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: getTactic()");
        Tactic tac = null;
        try {
            tac = calc_tree_.getMathEngine().getTactic(calc_tree_id_, sml_pos_);
        } catch (RemoteException e) {
            e.printStackTrace();
        }

        if (tac != null)
            if (logger.isInfoEnabled())
                logger.info("  DG<-BR: getTactic <-" + tac.toSMLString());

        return tac;
    }

    /**
     * @see isac.interfaces.ICalcIterator#getApplicableTactics()
     */
    public Vector getApplicableTactics(int scope) throws RemoteException {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: getApplicableTactics()");
        Vector tacs = null;
        try {
            tacs = calc_tree_.getMathEngine().getApplicableTactics(
                    calc_tree_id_, scope, sml_pos_);
        } catch (RemoteException e) {

            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: getApplicableTactics()");
        return tacs;
    }

    /**
     * get the Assumptions generated by the Tactic applied to the Formula at a
     * CalcIterator-position
     * 
     * @see isac.interfaces.ICalcIterator#getAssumptions()
     */
    public Assumptions getAssumptions() {
        if (logger.isInfoEnabled())
            logger.info("  DG->BR: getAssumptions()");
        Assumptions asms = null;
        try {
            asms = calc_tree_.getMathEngine().getAssumptions(calc_tree_id_,
                    sml_pos_);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: getAssumption() <- " + asms.toSMLString());
        return asms;
    }

    /**
     * @see isac.interfaces.ICalcIterator#getAccumulatedAssumptions()
     */
    public Assumptions getAccumulatedAssumptions() throws RemoteException {
        Assumptions asms = null;
        try {
            asms = calc_tree_.getMathEngine().getAccumulatedAssumptions(
                    calc_tree_id_, sml_pos_);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: getAssumption() <- " + asms.toSMLString());
        return asms;
    }

    /**
     * @see isac.interfaces.ICalcIterator#tryMatch()
     */
    public Match initMatchProblem() throws RemoteException {
        Match m = null;
        try {
            m = calc_tree_.getMathEngine().initMatchProblem(calc_tree_id_);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: tryMatchProblem() <- " + m.toString());
        return m;
    }

    /**
     * @see isac.interfaces.ICalcIterator#tryMatch()
     */
    public Match tryMatchProblem(ProblemID problemID) throws RemoteException {
        Match m = null;
        try {
            m = calc_tree_.getMathEngine().tryMatchProblem(calc_tree_id_,
                    problemID);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: tryMatchProblem() <- " + m.toString());
        return m;
    }

    /**
     * @see isac.interfaces.ICalcIterator#tryRefine()
     */
    public Match tryRefineProblem(ProblemID problemID) throws RemoteException {
        Match m = null;
        try {
            m = calc_tree_.getMathEngine().tryRefineProblem(calc_tree_id_,
                    problemID);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: tryRefineProblem() <- " + m.toString());
        return m;
    }

    /**
     * @see isac.interfaces.ICalcIterator#tryMatchMethod(HierarchyKey)
     */
    public Match initMatchMethod() throws RemoteException {
        Match m = null;
        try {
            m = calc_tree_.getMathEngine().initMatchMethod(calc_tree_id_);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: initMatchMethod() <- " + m.toString());
        return m;
    }

    /**
     * @see isac.interfaces.ICalcIterator#tryMatchMethod(HierarchyKey)
     */
    public Match tryMatchMethod(MethodID methodID) throws RemoteException {
        Match m = null;
        try {
            m = calc_tree_.getMathEngine().tryMatchMethod(calc_tree_id_,
                    methodID);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        if (logger.isInfoEnabled())
            logger.info("  DG<-BR: tryMatchMethod() <- " + m.toString());
        return m;
    }

    public Context initContext(ContextType type) throws RemoteException {
        Context context = calc_tree_.getMathEngine().initContext(calc_tree_id_,
                type, sml_pos_);
        return context;
    }

    public Context checkContext(Context context2check) throws RemoteException {
        Context context = calc_tree_.getMathEngine().checkContext(
                calc_tree_id_, context2check, sml_pos_);
        return context;
    }

    public Context refineProblem(Context context2refine) throws RemoteException {
        Context context = calc_tree_.getMathEngine().refineProblem(
                calc_tree_id_, context2refine, sml_pos_);
        return context;
    }

    public CalcHead setContext(ContextProblem context2set)
            throws RemoteException {
        return calc_tree_.setContext(calc_tree_id_, context2set, sml_pos_);
    }

    public CalcHead setContext(ContextMethod context2set)
            throws RemoteException {
        return calc_tree_.setContext(calc_tree_id_, context2set, sml_pos_);
    }

    public void setContext(ContextTheory context2set) throws RemoteException {
        calc_tree_.setContext(calc_tree_id_, context2set, sml_pos_);
    }

    public String toSMLString() {
        if (sml_pos_ == null)
            return "([],Und)";
        else
            return sml_pos_.toSMLString();
    }

}